﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Linq;
using System.Reflection;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Scripting;

namespace HeuristicLab.Problems.Programmable {
  [Item("ProgrammableProblemScript", "Script that defines the parameter vector and evaluates the solution.")]
  [StorableClass]
  public class ProgrammableProblemScript : Script, IStorableContent {
    public string Filename { get; set; }
    protected bool SuppressEvents { get; set; }

    [StorableConstructor]
    protected ProgrammableProblemScript(bool deserializing) : base(deserializing) { }
    protected ProgrammableProblemScript(ProgrammableProblemScript original, Cloner cloner)
      : base(original, cloner) { }
    public ProgrammableProblemScript() { }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new ProgrammableProblemScript(this, cloner);
    }

    private volatile ProgrammableProblemBase instance;
    private object locker = new object();
    public ProgrammableProblemBase Instance {
      get {
        SuppressEvents = true;
        try {
          var oldInstance = instance;
          var compilationNecessary = false;
          if (instance == null) {
            lock (locker) {
              if (instance == null) {
                compilationNecessary = true;
                Compile();
              }
            }
          }
          if (compilationNecessary && (oldInstance != null || instance != null))
            OnInstanceChanged();
          return instance;
        } finally {
          SuppressEvents = false;
        }
      }
      protected set {
        instance = value;
        OnInstanceChanged();
      }
    }

    protected override string CodeTemplate {
      get {
        return @"using System;
using System.Linq;
using System.Collections.Generic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.ParameterVectorEncoding;
using HeuristicLab.Encodings.PermutationEncoding;

public class MyProblem : HeuristicLab.Problems.Programmable.ProgrammableProblemBase {
  public MyProblem() {
    // initialize private fields
  }
  public override bool IsMaximizationProblem { get { return false; } }
  public override ParameterVector GetParametersToOptimize() {
    return new ParameterVectorBuilder()
      // .AddBoolean(""Parameter"")
      // .AddInteger(""Parameter"", min: 5, max: 15, step: 3)
      // .AddReal(""Parameter"", min: 0.0, max: 1.0)
      // .SetPermutation(type: PermutationTypes.Absolute, length: 10)
    .Build();
  }

  public override void EvaluateFeasibility(ParameterVector vector) {
    vector.IsFeasible = true
    // && vector.Integer(""Parameter1"") < vector.Integer(""Parameter2"");
    // && (vector.Real(""Parameter3"") >= 1.0 || (vector.Real(""Parameter3"") < 1.0 && vector.Integer(""Parameter1"") < 5));
    ;
  }

  public override double Evaluate(IRandom random, ParameterVector vector) {
    var quality = 0.0;
    return quality;
  }

  // implement further classes and methods
}";
      }
    }

    public override Assembly Compile() {
      var assembly = base.Compile();
      var types = assembly.GetTypes();
      try {
        instance = (ProgrammableProblemBase)Activator.CreateInstance(types.First(x => typeof(ProgrammableProblemBase).IsAssignableFrom(x)));
      } catch {
        instance = null;
      }
      if (!SuppressEvents) OnInstanceChanged();
      return assembly;
    }

    protected override void OnCodeChanged() {
      base.OnCodeChanged();
      instance = null;
    }

    public event EventHandler InstanceChanged;
    protected void OnInstanceChanged() {
      var handler = InstanceChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }
  }
}
